/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
// TODO
// * Fermat interface.
// * Adaptive weight bit choice.
// * Remove m_trivialsectors and m_sector_weight from Topology?


#include <cstddef>
#include <utility>
#include <unordered_map>
#include <algorithm> // min, max, swap
#include <numeric> // accumulate
#include <fstream>
#include <array>
#include <tuple>

#include "integrals.h"

namespace pyred {

/****************
 * Compositions *
 ****************/

int binomial(int n, int k) {
  // Binomial coefficient 'n choose k'.
  if (k < 0 || k > n) return 0;
  int bino = 1;
  for (int i = 0; i != std::min(k, n-k); ++i) {
    bino = (bino * (n-i)) / (i+1);
  }
  return bino;
}

std::vector<std::vector<uint32_t>> compositions(uint32_t n, uint32_t k) {
  // Return a list of all compositions of the integer n into k parts.
  std::vector<std::vector<uint32_t>> compos;
  std::vector<std::vector<uint32_t>> tmp;
  compos.reserve(binomial(n+k-1,k-1));
  tmp.reserve(compos.size());
  compos.emplace_back();
  // stop condition i<k (not i!=k) to cover the case k=0.
  for (uint32_t i = 1; i < k; ++i) {
    for (const auto &c: compos) {
      auto clen = std::accumulate(c.cbegin(), c.cend(),
                                  static_cast<uint32_t>(0));
      for (uint32_t j = 0; j != n+1-clen; ++j) {
        tmp.emplace_back(c);
        tmp.back().emplace_back(j);
      }
    }
    compos.swap(tmp);
    tmp.clear();
  }
  if (k > 0) {
    for (auto &c: compos) {
      c.emplace_back(n-std::accumulate(c.cbegin(), c.cend(), 0));
    }
  }
  return compos;
}

// Cache for get_composition_id().
// Note that pointers/references to values in an unordered_map
// remain valid after insert/erase/rehashing.
std::unordered_map<
  std::pair<uint32_t,uint32_t>,
  std::pair<
    std::unordered_map<std::vector<uint32_t>,uint32_t>,
    std::vector<std::vector<uint32_t>>
  >
> composition_id;


const std::pair<
  std::unordered_map<std::vector<uint32_t>,uint32_t>,
  std::vector<std::vector<uint32_t>>
>& get_composition_id(uint32_t dots, uint32_t lines) {
  // Return the composition weight dictionary 'compo_weights'
  // and the composition from weight list 'compo_from_weight'.
  // compo_weights[compo] = weight and compo_from_weight[weight] = compo
  // for a composition of the integer 'dots' into 'lines' parts.
  static std::mutex mtx;
  {
    std::lock_guard<std::mutex> lock(mtx);
    auto resit = composition_id.find({dots, lines});
    if (resit != composition_id.cend()) {
      return resit->second;
    }
  }
  std::unordered_map<std::vector<uint32_t>,uint32_t> compo_weights;
  std::vector<std::vector<uint32_t>> compo_from_weight;
  uint32_t weight{0};
  auto compos = compositions(dots,lines);
  for (const auto &compo: compos) {
    compo_weights.insert({compo, weight});
    compo_from_weight.push_back(compo);
    ++weight;
  }
  std::lock_guard<std::mutex> lock(mtx);
  // Check if the result has been added to the cache in the meantime.
  // If yes, return the result from the cache to not invalidate
  // references to it.
  auto resit = composition_id.find({dots, lines});
  if (resit != composition_id.cend()) {
    return resit->second;
  }
  // Otherwise add it to the cache and return.
  auto it = composition_id.emplace(
    std::make_pair(dots, lines),
    std::make_pair(compo_weights, compo_from_weight)
  ).first;
  return it->second;
}

/******************
 * Seed generator *
 ******************/

Seeds::Seeds(uint32_t np, uint32_t sect, uint32_t dots, uint32_t sps)
: sector(np, false)
{
  uint32_t lines{0};
  uint32_t nums;
  for (auto pr = sector.begin(); pr != sector.end(); ++pr) {
    if (sect & 1) {
      ++lines;
      *pr = true;
    }
    sect >>= 1;
  }
  nums = np - lines;
  const auto& ppows = get_composition_id(dots,lines).second;
  const auto& spows = get_composition_id(sps,nums).second;
  ppows_begin = ppows.crbegin();
  ppows_end = ppows.crend();
  spows_begin = spows.cbegin();
  spows_end = spows.cend();
}

SeedIterator::SeedIterator()
: sector{}, p_begin{}, p_it{}, p_end{}, s_it{}, s_end{}
{}

SeedIterator::SeedIterator(
  const std::vector<bool>& sector,
  const compos_rit_type& p_begin, const compos_rit_type& p_end,
  const compos_it_type& s_begin, const compos_it_type& s_end)
: sector{sector},
  p_begin{p_begin}, p_it{p_begin}, p_end{p_end}, s_it{s_begin}, s_end{s_end}
{}

SeedIterator& SeedIterator::operator++() {
  if (p_it != p_end) {
    ++p_it;
  }
  if (p_it == p_end) {
    p_it = p_begin;
    ++s_it;
  }
  if (s_it == s_end) {
    // The iterator is depleted.
    sector.clear();
    p_begin = compos_rit_type{};
    p_it = compos_rit_type{};
    p_end = compos_rit_type{};
    s_it = compos_it_type{};
    s_end = compos_it_type{};
  }
  return *this;
}

SeedIterator SeedIterator::operator++(int) {
  auto it = *this;
  ++(*this);
  return it;
}

std::vector<pow_type> SeedIterator::operator*() const {
  std::vector<pow_type> powers;
  powers.reserve(sector.size());
  auto ppows_it = p_it->cbegin();
  auto spows_it = s_it->cbegin();
  for (bool pr: sector) {
    if (pr) {
      powers.push_back(1+(*(ppows_it++)));
    }
    else {
      powers.push_back(-(*(spows_it++)));
    }
  }
  return powers;
}

std::unique_ptr<std::vector<pow_type>> SeedIterator::operator->() const {
  std::unique_ptr<std::vector<pow_type>> tmp{new std::vector<pow_type>};
  *tmp = *(*this);
  return tmp;
}

bool operator==(const SeedIterator& lhs, const SeedIterator& rhs) {
  return (lhs.sector == rhs.sector && lhs.p_begin == rhs.p_begin &&
          lhs.p_it == rhs.p_it && lhs.p_end == rhs.p_end &&
          lhs.s_it == rhs.s_it && lhs.s_end == lhs.s_end);
}

bool operator!=(const SeedIterator& lhs, const SeedIterator& rhs) {
  return !(lhs == rhs);
}

void swap(SeedIterator& lhs, SeedIterator& rhs) {
  std::swap(lhs.sector, rhs.sector);
  std::swap(lhs.p_begin, rhs.p_begin);
  std::swap(lhs.p_it, rhs.p_it);
  std::swap(lhs.p_end, rhs.p_end);
  std::swap(lhs.s_it, rhs.s_it);
  std::swap(lhs.s_end, lhs.s_end);
}

/******************
 * Integral class *
 ******************/

// static members
uint32_t Integral::s_np{0};
uint32_t Integral::s_ordering{0};
std::array<uint32_t,6> Integral::s_weight_bits{{0,0,5,4,17,13}};
std::array<weight_type,6> Integral::s_weight_proj{{0,0,0,0,0,0}};
uint32_t Integral::s_minlines{1u<<31};
int Integral::s_cache_level{0};
std::unordered_map<Integral,weight_type> Integral::s_i2w_cache{};
std::unordered_map<weight_type,Integral> Integral::s_w2i_cache{};
std::mutex Integral::s_i2w_mtx{};
std::mutex Integral::s_w2i_mtx{};
std::vector<std::unordered_set<uint32_t>> Integral::s_trivialsectors{};
std::vector<std::vector<std::pair<uint32_t,uint32_t>>>
  Integral::s_sector_weight{};

Integral::Integral(uint32_t topo, const std::vector<pow_type>& pows)
: m_topoid{topo}
, m_powers{pows}
{
  if (size() != s_np) {
    throw init_error("Integral(): wrong number of indices");
  }
}

Integral::Integral(uint32_t topo, std::vector<pow_type>&& pows)
: m_topoid{topo}
, m_powers{std::move(pows)}
{
  if (size() != s_np) {
    throw init_error("Integral(): wrong number of indices");
  }
}

Integral::Integral(std::istringstream& ss, uint32_t topo)
: m_topoid{topo}
{
  // Input format: "*[n1,n2,..]*"
  // Skip beginning until '[' is found, stop at ']'.
  reserve(s_np);
  pow_type pow;
  char sep;
  do {
    ss >> sep;
  } while (sep != '[');
  sep = ',';
  while (sep == ',') {
    ss >> pow;
    m_powers.push_back(pow);
    ss >> sep;
  }
  if (sep != ']') {
    throw init_error("Integral parser: ill-formatted integral");
  }
  if (size() != s_np) {
    throw init_error("Integral parser: wrong number of indices");
  }
}

Integral::Integral(weight_type w) {
  // Convert an integer integral weight to an integral.
  if (!w) {
    throw input_error(
      "Integral(): cannot construct an integral of zero weight.");
  }
  if (s_cache_level & 1) {
    std::lock_guard<std::mutex> lock(s_w2i_mtx);
    auto posit = s_w2i_cache.find(w);
    if (posit != s_w2i_cache.cend()) {
      *this = posit->second;
      return;
    }
  }
  auto props = properties(w);
  m_topoid = props.topology;
  m_powers.reserve(s_np);
  // * dots to the left have lower weight
  // auto ppows_it = get_composition_id(props.dots,props.lines).second.cend()[
  //                   -1-static_cast<int>(props.ppows_weight)].cbegin();
  // * dots to the right have lower weight
  auto ppows_it = get_composition_id(props.dots,props.lines).second[
                    props.ppows_weight].cbegin();
  // * sps to the right have lower weight
  // auto spows_it = get_composition_id(props.sps,props.nums).second[
  //                   props.spows_weight].cbegin();
  // * sps to the left have lower weight
  auto spows_it = get_composition_id(props.sps,props.nums).second.cend()[
                    -1-static_cast<int>(props.spows_weight)].cbegin();
  for (uint32_t dummy = 0; dummy != s_np; ++dummy) {
    if (props.sector & 1) {
      m_powers.push_back(1+(*(ppows_it++)));
    }
    else {
      m_powers.push_back(-(*(spows_it++)));
    }
    props.sector >>= 1;
  }
  // Insert the element into the cache(s)
  // if it hasn't been cached in the meantime.
  if (s_cache_level & 2) {
    std::lock_guard<std::mutex> lock(s_i2w_mtx);
    s_i2w_cache.insert({*this, w});
  }
  if (s_cache_level & 1) {
    std::lock_guard<std::mutex> lock(s_w2i_mtx);
    s_w2i_cache.insert({w, *this});
  }
}

IntegralProperties Integral::properties(weight_type w) {
  // Extract topology, sector, number of lines, numerators,
  // dots, scalar products, dot and sp composition weight.
  // Return them as an IntegralProperties instance.
  // * dot composition has higher weight than sp composition
  // uint32_t spow_w = w & s_weight_proj[5];
  // w >>= s_weight_bits[5];
  // uint32_t ppow_w = w & s_weight_proj[4];
  // w >>= s_weight_bits[4];
  // * sp composition has higher weight than dot composition
  uint32_t ppow_w = w & s_weight_proj[4];
  w >>= s_weight_bits[4];
  uint32_t spow_w = w & s_weight_proj[5];
  w >>= s_weight_bits[5];
  uint32_t dotsp_weight_2 = w & s_weight_proj[3];
  w >>= s_weight_bits[3];
  uint32_t dotsp_weight_1 = w & s_weight_proj[2];
  w >>= s_weight_bits[2];
  auto wforsect = w;
  w >>= s_weight_bits[1];
  auto topoid = static_cast<uint32_t>(w);
  uint32_t sect = s_sector_weight.at(topoid)[wforsect & s_weight_proj[1]].second;
  uint32_t sectcpy{sect};
  uint32_t lines{0};
  for (uint32_t dummy = 0; dummy != s_np; ++dummy) {
    if (sectcpy & 1) ++lines;
    sectcpy >>= 1;
  }
  uint32_t nums = s_np - lines;
  uint32_t dots;
  uint32_t sps;
  std::tie(dots, sps) = weight_to_dots(dotsp_weight_1, dotsp_weight_2, lines);
  return IntegralProperties{topoid, sect, lines, nums,
                            dots, sps, ppow_w, spow_w};
}

std::pair<uint32_t,uint32_t> Integral::weight_to_dots(
    uint32_t dotsp_1, uint32_t dotsp_2, uint32_t lines) {
  uint32_t dots{0};
  uint32_t sps{0};
  switch (s_ordering) {
    case 1:
      dots = dotsp_1;
      sps = dotsp_2;
      break;
    case 2:
      sps = dotsp_1;
      dots = dotsp_2;
      break;
    case 3:
      dots = dotsp_2;
      sps = dotsp_1 - dotsp_2;
      break;
    case 4:
      dots = dotsp_1 - dotsp_2;
      sps = dotsp_2;
      break;
    case 5:
      dots = dotsp_1 - (lines - s_minlines);
      sps = dotsp_2;
      break;
  }
  return {dots, sps};
}

void Integral::setup(uint32_t ordering,
                     const std::vector<uint32_t>& weight_bits) {
  static std::mutex setup_mtx;
  std::lock_guard<std::mutex> lock(setup_mtx);
  if (Topology::s_next_id) {
    throw init_error("Integral::setup() can only be called"
                     "before creating Topology objects.");
  }
  if (ordering < 1 || ordering > 5) {
    throw init_error("Integral::setup(): ordering must be one of {1,2,3,4,5}");
  }
  s_ordering = ordering;
  if (!weight_bits.empty()) {
    if (weight_bits.size() != s_weight_bits.size() - 2) {
      std::ostringstream ss;
      ss << "Integral::setup(): weight_bits must have "
         << s_weight_bits.size() - 2 << " elements.";
      throw init_error(ss.str());
    }
    // s_weight_bits[0] and s_weight_bits[1] are assigned when the first
    // Topology object is created (i.e. when np is known).
    for (std::size_t n = 2; n != 6; ++n) {
      // weights for dotsp_1, dotsp_2 (meaning depending on ordering),
      // dot distribution, sp distribution
      s_weight_bits[n] = weight_bits[n-2];
    }
  }
}

void Integral::assign_weight_bits() {
  uint32_t nontopo_weight = s_np + s_weight_bits[2] + s_weight_bits[3]
                                 + s_weight_bits[4] + s_weight_bits[5];
  if (nontopo_weight > 64) {
    throw init_error("Integral weight representation exceeds 64 bits.");
  }
  s_weight_bits[0] = 64 - nontopo_weight; // topology weight: the rest
  s_weight_bits[1] = s_np; // sector weight
  for (std::size_t n = 0; n != 6; ++n) {
    s_weight_proj[n] = (1 << s_weight_bits[n]) - 1;
  }
}

Seeds Integral::seeds(uint32_t sect, uint32_t dots, uint32_t sps) {
  return Seeds(s_np, sect, dots, sps);
}

weight_type Integral::to_weight() const {
  // Convert an integral to an integer weight.
  if (s_cache_level & 2) {
    std::lock_guard<std::mutex> lock(s_i2w_mtx);
    auto posit = s_i2w_cache.find(*this);
    if (posit != s_i2w_cache.cend()) {
      return posit->second;
    }
  }
  weight_type weight{m_topoid};
  uint32_t lines{0};
  uint32_t dots{0};
  uint32_t nums{0};
  uint32_t sps{0};
  uint32_t dotsp_weight_1{0};
  uint32_t dotsp_weight_2{0};
  std::vector<uint32_t> ppows;
  std::vector<uint32_t> spows;
  uint32_t sector{0};
  uint32_t linepos{0};
  for (const auto p: m_powers) {
    if (p > 0) {
      sector += (1 << linepos);
      ++lines;
      dots += (p-1);
      ppows.push_back(p-1);
    }
    else {
      sps -= p;
      spows.push_back(-p);
    }
    ++linepos;
  }
  if (s_trivialsectors.at(m_topoid).find(sector) !=
      s_trivialsectors.at(m_topoid).end()) {
    return 0;
  }
  weight <<= s_weight_bits[1];
  weight += s_sector_weight[m_topoid][sector].first;
  nums = s_np - lines;
  // dot and sp composition weights
  // * dots to the left have lower weight
  // const auto& cid = get_composition_id(dots,lines);
  // auto ppows_weight = cid.second.size() - cid.first.at(ppows) - 1;
  // * dots to the right have lower weight
  auto ppows_weight = get_composition_id(dots,lines).first.at(ppows);
  // * sps to the right have lower weight
  // auto spows_weight = get_composition_id(sps,nums).first.at(spows);
  // * sps to the left have lower weight
  const auto& cid = get_composition_id(sps,nums);
  auto spows_weight = cid.second.size() - cid.first.at(spows) - 1;
  // (dots, sps) ordering
  switch (s_ordering) {
    case 1:
      dotsp_weight_1 = dots;
      dotsp_weight_2 = sps;
      break;
    case 2:
      dotsp_weight_1 = sps;
      dotsp_weight_2 = dots;
      break;
    case 3:
      dotsp_weight_1 = dots + sps;
      dotsp_weight_2 = dots;
      break;
    case 4:
      dotsp_weight_1 = dots + sps;
      dotsp_weight_2 = sps;
      break;
    case 5:
      dotsp_weight_1 = dots + (lines - s_minlines);
      dotsp_weight_2 = sps;
      break;
  }
  (weight <<= s_weight_bits[2]) += dotsp_weight_1;
  (weight <<= s_weight_bits[3]) += dotsp_weight_2;
  // * dot composition has higher weight than sp composition
  // (weight <<= s_weight_bits[4]) += ppows_weight;
  // (weight <<= s_weight_bits[5]) += spows_weight;
  // * sp composition has higher weight than dot composition
  (weight <<= s_weight_bits[5]) += spows_weight;
  (weight <<= s_weight_bits[4]) += ppows_weight;
  if (m_topoid > s_weight_proj[0] ||
      dotsp_weight_1 > s_weight_proj[2] || dotsp_weight_2 > s_weight_proj[3] ||
      ppows_weight > s_weight_proj[4] || spows_weight > s_weight_proj[5]) {
    weight_overflow(m_topoid, dotsp_weight_1, dotsp_weight_2,
                    ppows_weight, spows_weight, lines);
  }
  // Insert the element into the cache(s)
  // if it hasn't been cached in the meantime.
  if (s_cache_level & 2) {
    std::lock_guard<std::mutex> lock(s_i2w_mtx);
    s_i2w_cache.insert({*this, weight});
  }
  if (s_cache_level & 1 && weight) {
    // Do not cache weight=0; constructing Integral(0) is not allowed.
    std::lock_guard<std::mutex> lock(s_w2i_mtx);
    s_w2i_cache.insert({weight, *this});
  }
  return weight;
}

void Integral::weight_overflow(
    uint32_t topo, uint32_t dotsp_1, uint32_t dotsp_2,
    uint32_t ppow_w, uint32_t spow_w, uint32_t lines) {
  // Error handling in case a sub-weight does not fit
  // in the Integral weight representation, because too few bits
  // are reserved for the particular sub-weight.
  std::ostringstream ss;
  ss << "Integral weight overflow in";
  uint32_t dots;
  uint32_t sps;
  std::tie(dots, sps) = weight_to_dots(dotsp_1, dotsp_2, lines);
  if (topo > s_weight_proj[0]) {
    ss << " topology (" << topo
       << " does not fit in " << s_weight_bits[0] << " bits);";
  }
  if (dotsp_1 > s_weight_proj[2]) {
    switch (s_ordering) {
      case 1:
        ss << " dots (" << dots
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
      case 2:
        ss << " sps (" << sps
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
      case 3:
      case 4:
        ss << " dots+sps (" << dots << "+" << sps
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
      case 5:
        ss << " dots+lines-minlines ("
           << dots << "+" << lines << "-" << s_minlines
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
    }
  }
  if (dotsp_2 > s_weight_proj[3]) {
    switch (s_ordering) {
      case 1:
      case 4:
      case 5:
        ss << " sps (" << sps
           << " does not fit in " << s_weight_bits[3] << " bits);";
        break;
      case 2:
      case 3:
        ss << " dots (" << dots
           << " does not fit in " << s_weight_bits[4] << " bits);";
        break;
    }
  }
  if (ppow_w > s_weight_proj[4]) {
    ss << " dot distribution (" << ppow_w
       << " does not fit in " << s_weight_bits[4] << " bits);";
  }
  if (spow_w > s_weight_proj[5]) {
    ss << " sp distribution (" << spow_w <<
          " does not fit in " << s_weight_bits[5] << " bits);";
  }
  ss << "\nuse Integral::setup() to reserve more space where needed.";
  throw init_error(ss.str());
}

Integral Integral::zip(const std::vector<pow_type>& merge) const {
  // Return a new Integral with 'm_powers' and 'merge' added componentwise.
  std::vector<pow_type> pows;
  pows.reserve(s_np);
  auto p1 = m_powers.cbegin();
  for (auto p2: merge) {
    pows.push_back(*p1+p2);
    ++p1;
  }
  return Integral(m_topoid, std::move(pows));
}

void Integral::use_cache(int level) {
  s_cache_level = level;
}

void Integral::clear_cache(int level) {
  if (level & 1) {
    s_w2i_cache.clear();
  }
  if (level & 2) {
    s_i2w_cache.clear();
  }
}

bool Integral::operator<(const Integral& rhs) const {
  return (to_weight() < rhs.to_weight());
}

bool operator==(const Integral& lhs, const Integral& rhs) {
  return (lhs.m_topoid == rhs.m_topoid && lhs.m_powers == rhs.m_powers);
}

bool operator!=(const Integral& lhs, const Integral& rhs) {
  return (lhs.m_topoid != rhs.m_topoid || lhs.m_powers != rhs.m_powers);
}

std::ostream& operator<<(std::ostream& out, const Integral& igl) {
  out << Topology::id_to_topo(igl.m_topoid).m_name << igl.m_powers;
  return out;
}

/******************
 * Topology class *
 ******************/

// static members
uint32_t Topology::s_next_id{0};
std::vector<std::unique_ptr<Topology>> Topology::s_id_to_topo{};

Topology::Topology(const std::string& name, const uint32_t np,
                   const std::vector<uint32_t>& trivialsectors)
: m_name{name}
, m_np{np}
, m_id{s_next_id}
, m_trivialsectors{trivialsectors.cbegin(), trivialsectors.cend()} {
  if (!m_np) {
    throw init_error("Topology(): np=0 is not allowed.");
  }
  if (Integral::s_np != m_np) {
    if (Integral::s_np) {
      throw init_error("Topology(): all topologies must have the same np.");
    }
    else {
      // This is the first topology which is created.
      Integral::s_np = m_np;
      Integral::assign_weight_bits();
    }
  }
  ++s_next_id;
  assign_sector_weight();
  Integral::s_trivialsectors.push_back(m_trivialsectors);
  Integral::s_sector_weight.push_back(m_sector_weight);
}

Topology& new_topology(const std::string& name, const uint32_t np,
                       const std::vector<uint32_t>& trivialsectors) {
  Topology::s_id_to_topo.push_back(
    std::unique_ptr<Topology>{new Topology{name, np, trivialsectors}});
  return *Topology::s_id_to_topo.back();
}

Topology& new_topology(const std::string& name, const std::string& projectdir) {
  // Read the trivial sectors from a Kira file.
  // Format: "sn1,sn2,..."
  std::vector<uint32_t> trivialsectors;
  {
    std::string tsfile = projectdir + "/sectormappings/"
                      + name + "/trivialsector";
    std::ifstream instream{tsfile};
    if (!instream.good()) {
      std::ostringstream ss;
      ss << "Failed reading trivial sector file " << tsfile;
      throw init_error(ss.str());
    }
    std::string line;
    std::getline(instream, line);
    std::istringstream ss{line};
    uint32_t sn;
    char sep;
    do {
      ss >> sn;
      trivialsectors.push_back(sn);
    } while (static_cast<bool>(ss >> sep));
    // TODO: error handling for invalid file formats
  }
  // Read the non-trivial sectors from a Kira file just to get np.
  // Format: "nts1 lines1\nnts2 lines2\n..."
  uint32_t np{0};
  {
    std::string ntsfile = projectdir + "/sectormappings/"
                        + name + "/nonTrivialSector";
    std::ifstream instream{ntsfile};
    if (!instream.good()) {
      std::ostringstream ss;
      ss << "Failed reading non-trivial sector file " << ntsfile;
      throw init_error(ss.str());
    }
    std::string line;
    while (std::getline(instream, line)) {
      uint32_t sectn;
      uint32_t sectnp;
      std::istringstream ss{line};
      if (!static_cast<bool>(ss >> sectn >> sectnp)) {
        std::ostringstream err;
        err << "invalid data in non-trivial sector file " << ntsfile
            << ": \"" << line << "\"";
        throw init_error(err.str());
      }
      np = std::max(np, sectnp);
    }
  }
  return new_topology(name, np, trivialsectors);
}

void Topology::assign_sector_weight() {
  // assign m_sector_weight and m_minlines
  // sector weight table
  //   m_sector_weight[sector].first = weight
  //   m_sector_weight[weight].second = sector
  std::vector<std::array<uint32_t,3>> sect_w;
  m_minlines = m_np;
  for (uint32_t s = 0; s != (1u << m_np); ++s) {
    uint32_t scpy{s};
    uint32_t lines{0};
    uint32_t lines_w{0};
    for (uint32_t l = 0; l != m_np; ++l) {
      if (scpy & 1) {
        ++lines;
      }
      else {
        ++lines_w;
      }
      scpy >>= 1;
      lines_w <<= 1;
    }
    if (Integral::s_ordering != 5) {
      sect_w.push_back({{s, lines, lines_w}});
    }
    else {
      sect_w.push_back({{s, 0, s}});
    }
    // m_minlines: minimal number of lines in any non-trivial sector.
    if (lines < m_minlines &&
        m_trivialsectors.find(s) == m_trivialsectors.end()) {
      m_minlines = lines;
    }
  }
  Integral::s_minlines = std::min(Integral::s_minlines, m_minlines);
  // sort by (number of lines, weight for fixed lines) --> global weight
  std::sort(sect_w.begin(), sect_w.end(),
    [](const std::array<uint32_t,3>& slw1, const std::array<uint32_t,3>& slw2) {
      return slw1[1] == slw2[1] ? slw1[2] < slw2[2] : slw1[1] < slw2[1];
    }
  );
  // assign sector weight and m_sector_weight[weight].second = sector
  for (uint32_t sw = 0; sw != sect_w.size(); ++sw) {
    sect_w[sw][2] = sw;
    m_sector_weight.emplace_back(0, sect_w[sw][0]);
  }
  // sort by sector number
  std::sort(sect_w.begin(), sect_w.end(),
    [](const std::array<uint32_t,3>& slw1, const std::array<uint32_t,3>& slw2) {
      return slw1[0] < slw2[0];
    }
  );
  // assign m_sector_weight[sector].first = weight
  for (uint32_t sw = 0; sw != sect_w.size(); ++sw) {
    m_sector_weight[sw].first = sect_w[sw][2];
  }
}

}
