/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef PYRED_INTERFACE_H
#define PYRED_INTERFACE_H

#include <iostream>
#include <iomanip>
#include <cstddef>
#include <utility>
#include <vector>
#include <unordered_set>
#include <unordered_map>
#include <tuple>
#include <algorithm>
#include <string>
#include <sstream>
#include <ctime>
#include <limits>

#include "coeff_helper.h"
#include "coeff_int.h"
#include "coeff_vec.h"
#include "gauss.h"

namespace pyred {

using content_type = std::unordered_map<intid,
  std::tuple<intid,std::vector<intid>,std::vector<intid>>>;

float time_diff(std::clock_t, std::clock_t);


class Logger {
  // Logger log{verbosity};
  // log(level) << out;
  // print 'out' if level <= verbosity.
private:
  int verbosity_val;
  int level;
public:
  Logger(int = 1, int = 1);
  void set_verbosity(int);
  int verbosity();
  Logger& operator()(int);
  template<typename T>
  Logger& operator<<(T&& out) {
    if (level <= verbosity_val) std::cout << std::forward<T>(out);
    return *this;
  }
  // Specialisation for I/O manipulators std::endl and std::flush.
  Logger& operator<<(std::ostream& (*)(std::ostream&));
};


class Config
{
public:
  static void set_coeff_cls(int);
  static void set_solver_type(int);
  static void set_backward(bool);
  static void set_verbosity(int);
  static int coeff_cls();
  static int solver_type();
  static bool backward();
  static int verbosity();
  static Logger& log(int);
  static std::pair<std::string,std::string> parse(const std::string& = "");
  static std::pair<std::string,std::string> parse(
    const std::vector<std::string>&);
private:
  static Logger log_val;
  static int coeff_cls_val; // 1=Coeff_int, 2=Coeff_vec
  static int solver_type_val;
  static bool backward_val;
};


class System
{
public:
  System();
  System(const std::vector<eqdata>&);
  System(std::vector<eqdata>&&);
  System(const std::string&);
  std::size_t size() const;
  void reserve(const std::size_t);
  void add(const eqdata&);
  void add(eqdata&&);
  void add(const std::vector<eqdata>&);
  void add(std::vector<eqdata>&&);
  void add(const std::string&);
  void add_forward(const eqdata&,
                   intid neq = std::numeric_limits<intid>::max());
  std::vector<intid> solve();
  std::vector<intid> backward();
  const content_type& get_content() const;
  std::pair<std::vector<intid>,std::vector<intid>>
    select(const std::vector<intid>&, const std::vector<intid>& = {});
  std::vector<eqdata> retrieve();
  std::vector<eqdata> retrieve(const std::vector<intid>&);
private:
  std::vector<eqdata> sys;
  // unordered_map content[integral]
  // = tuple(eqnum, {unreduced_integrals}, {inserted_integrals})
  content_type content;
  template<typename Coeff>
  std::vector<intid> solve_tmpl();
  template<typename Coeff>
  std::vector<intid> backward_tmpl(SystemOfEqs<Coeff>&, int = 1);
  template<typename Coeff>
  void add_forward_tmpl(const eqdata&,
                        SystemOfEqs<Coeff>&,
                        sol_map<Coeff>&,
                        intid neq);
  // Data members for the on-the-fly forward solver.
  // For simplicity we use separate members for each template parameter.
  intid neqs = 0;
  SystemOfEqs<Coeff_int> numsys_int;
  SystemOfEqs<Coeff_vec> numsys_vec;
  sol_map<Coeff_int> sols_int;
  sol_map<Coeff_vec> sols_vec;
};


template<typename Coeff>
void System::add_forward_tmpl(const eqdata& eq,
                              SystemOfEqs<Coeff>& numsys,
                              sol_map<Coeff>& sols,
                              intid neq) {
  if (neq == std::numeric_limits<intid>::max()) {
    neq = neqs;
  }
  if (!(neqs++)) {
    numsys.sys.reserve(sys.capacity());
    sys.clear();
  }
  icpairs<Coeff> tmpeq;
  tmpeq.reserve(eq.size());
  for (const auto& ic: eq) {
    tmpeq.emplace_back(ic.first, parse_coeff<Coeff>(ic.second));
  }
  if (numsys.sys.size() == numsys.sys.capacity()) {
    throw init_error("pyred::System() grew beyond reserved capacity.");
  }
  numsys.sys.emplace_back(tmpeq, neq);
  auto& cureq = numsys.sys.back();
  cureq.solve(sols, true);
  if (!cureq.empty()) {
    sols.insert({cureq.front().first, std::ref(cureq)});
  }
  else {
    numsys.sys.pop_back();
  }
}

template<typename Coeff>
std::vector<intid> System::solve_tmpl() {
  auto time0 = std::clock();
  Config::log(1) << "parse coefficients" << std::flush;
  auto numsys = SystemOfEqs<Coeff>(sys);
  auto time1 = std::clock();
  auto nequations = numsys.size();
  Config::log(1) << ": " << time_diff(time0,time1)
            << " (" << nequations << " equations)" << std::endl;
  intid maxinsertions;
  decltype(numsys.independent()) indep_eqnums;
  std::size_t neqs_indep{0};
  std::size_t neqs_indep_chk{0};

  if (Config::solver_type() == 1) {
    // eliminator
    Config::log(1) << "solve forward" << std::flush;
    maxinsertions = numsys.solve();
    indep_eqnums = numsys.independent();
    neqs_indep = indep_eqnums.size();
    auto endtime = std::clock();
    Config::log(1) << ":      " << time_diff(time1,endtime) << std::endl;
  }
  else if (Config::solver_type() == 2) {
    // eliminator limit reconstruction + pivot(limit)
    intid limit{0};
    Config::log(1) << "solve forward 1/2" << std::flush;
    {
      auto numsys2 = numsys;
      numsys2.solve();
      // Reconstruction of the number of insertions.
      // Must be done before sorting the system again.
      // map: integral --> all inserted integrals
      std::unordered_map<intid,std::vector<intid>> allinserted_map;
      for (const auto& eq: numsys2.sys) {
        if (eq.empty()) continue;
        intid curint = eq.front().first;
        std::unordered_set<intid> allinserted_set{};
        for (const auto ins: eq.inserted) {
          allinserted_set.insert(ins);
          for (const auto recins: allinserted_map.find(ins)->second) {
            allinserted_set.insert(recins);
          }
        }
        auto& insvec = allinserted_map.emplace(
          curint, std::vector<intid>{}).first->second;
        for (const auto iid: allinserted_set) {
          // Count only forward insertions.
          if (iid > curint) insvec.push_back(iid);
        }
        limit = std::max(limit, static_cast<intid>(insvec.size()));
      }
      neqs_indep_chk = numsys2.independent().size();
    }
    auto endtime = std::clock();
    Config::log(1) << ":  " << time_diff(time1,endtime)
                   << " (limit: " << limit << ")" << std::endl;
    auto starttime = endtime;
    Config::log(1) << "solve forward 2/2" << std::flush;
    maxinsertions = numsys.psolve_map(limit);
    indep_eqnums = numsys.independent();
    neqs_indep = indep_eqnums.size();
    endtime = std::clock();
    Config::log(1) << ":  " << time_diff(starttime,endtime) << std::endl;
    if (neqs_indep_chk > neqs_indep) {
      std::ostringstream ss;
      ss << "Partial insertion eliminator requires less insertions"
         << " than pivot solver. This is unexpected and meas that the"
         << " settings are not optimal. Try again with the default solver."
         << std::endl;
      throw init_error(ss.str());
    }
  }
  else if (Config::solver_type() == 3) {
    // partial insertion + pivot(limit), check if ok
    intid limit;
    Config::log(1) << "solve forward 1/2" << std::flush;
    {
      auto numsys2 = numsys;
      limit = numsys2.solve(false); // partial insertion
      neqs_indep_chk = numsys2.independent().size();
    }
    auto endtime = std::clock();
    Config::log(1) << ":  " << time_diff(time1,endtime)
                   << " (limit: " << limit << ")" << std::endl;
    auto starttime = endtime;
    Config::log(1) << "solve forward 2/2" << std::flush;
    maxinsertions = numsys.psolve_map(limit);
    indep_eqnums = numsys.independent();
    neqs_indep = indep_eqnums.size();
    endtime = std::clock();
    Config::log(1) << ":  " << time_diff(starttime,endtime) << std::endl;
    if (neqs_indep_chk > neqs_indep) {
      std::ostringstream ss;
      ss << "Partial insertion eliminator requires less insertions"
         << " than pivot solver. This is unexpected and meas that the"
         << " settings are not optimal. Try again with the default solver."
         << std::endl;
      throw init_error(ss.str());
    }
  }
  else if (Config::solver_type() == 4) {
    // eliminator + pivot(limit*10), *2 until ok
    int solverun{0};
    intid limit;
    Config::log(1) << "solve forward " << ++solverun << "/?" << std::flush;
    {
      auto numsys2 = numsys;
      limit = numsys2.solve();
      limit *= 10;
      neqs_indep_chk = numsys2.independent().size();
    }
    auto endtime = std::clock();
    Config::log(1) << ":  " << time_diff(time1,endtime)
                   << " (limit: " << limit << ")" << std::endl;
    decltype(numsys) numsys2;
    while (neqs_indep < neqs_indep_chk) {
      Config::log(1) << "solve forward " << ++solverun << "/?" << std::flush;
      auto starttime = endtime;
      numsys2 = numsys;
      maxinsertions = numsys2.psolve_map(limit);
      indep_eqnums = numsys2.independent();
      neqs_indep = indep_eqnums.size();
      limit *= 2;
      endtime = std::clock();
      Config::log(1) << ":  " << time_diff(starttime,endtime) << std::endl;
    }
    numsys = numsys2;
  }
  else {
    std::ostringstream ss;
    ss << "Invalid solver type " << Config::solver_type() << std::endl;
    throw init_error(ss.str());
  }
  auto time2 = std::clock();
  Config::log(1) << "max insertions:     " << maxinsertions << std::endl;

  if (Config::backward()) {
    // backward insertion
    Config::log(1) << "solve backward" << std::flush;
    backward_tmpl<Coeff>(numsys);
    auto time3 = std::clock();
    Config::log(1) << ":     " << time_diff(time2,time3) << std::endl;
  }

  if (Config::backward() || Config::solver_type() != 1) {
    auto time4 = std::clock();
    Config::log(1) << "total solve time:   " << time_diff(time1, time4)
                   << std::endl;
  }
  Config::log(1) << std::setw(8) << nequations << " equations: "
                 << nequations-neqs_indep << " zero + "
                 << neqs_indep << " independent" << std::endl;
  return indep_eqnums;
}

template<typename Coeff>
std::vector<intid> System::backward_tmpl(SystemOfEqs<Coeff>& numsys,
                                         int mode) {
  std::vector<intid> indep_eqnums;
  if (mode != 1) {
    // independent() also sorts the system
    indep_eqnums = numsys.independent();
  }
  numsys.solve();
  // fill the unordered_map content: intid->(vector<intid>)
  for (const auto& eq: numsys.sys) {
    auto itchk = content.insert({
      eq.front().first,
      std::make_tuple(eq.eqnum, std::vector<intid>{}, std::vector<intid>{})
    }).first;
    auto& cont = std::get<1>(itchk->second);
    cont.reserve(eq.size()-1);
    // Empty equations have already been dropped by independent().
    for (auto it = eq.cbegin()+1; it != eq.cend(); ++it) {
      cont.emplace_back(it->first);
    }
    std::get<2>(itchk->second) = std::move(eq.inserted);
  }
  if (mode != 1) {
    Config::log(1) << std::setw(8) << neqs << " equations: "
                  << neqs-indep_eqnums.size() << " zero + "
                  << indep_eqnums.size() << " independent" << std::endl;
  }
  return indep_eqnums;
}

}

#endif
